;;; mw-api.el --- Interface to the MediaWiki API

;; Copyright (C) 2014 Free Software Foundation, Inc.

;; Author: Ivan Shmakov <ivan@siamics.net>
;; Version: 0.1
;; Keywords: web-services, comm

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;; This interface is inspired by mediawiki.el, but is intended to
;; provide an entirely independent implementation, sharing no code
;; with mediawiki.el whatsoever, and, while possibly similar at times,
;; not being actually compatible with it.

;; See also:
;;   https://www.mediawiki.org/wiki/API:Main_page
;;   https://launchpad.net/mediawiki-el

;;; Code:
(require 'cl-lib)			; for cl-ldiff
(require 'json)				; for json-read
(require 'mm-url)			; for mm-url-encode-www-form-urlencoded
(require 'url)				; for url-retrieve-synchronously
(require 'url-http)

(defvar mw-api-tokens (make-hash-table :test 'equal)
  "A hash of hashes of tokens seen in responses.")

(defun mw-api-get (keys alist)
  (while (and keys alist)
    (let ((k (car keys)))
      (setq alist
	    (cond ((not (or (eq t k) (eq nil k)))
		   (cdr (assoc k alist)))
		  ((or  (eq nil k) (not (cdr alist)))
		   (cdar alist))
		  (t (error "Single-element alist expected: %S" alist)))
	    keys  (cdr keys))))
  alist)

(defun mw-api-store-tokens (endpoint-uri tokens &optional predicate)
  (when tokens
    (unless predicate
      (setq predicate 'identity))
    (let ((hash (or (gethash endpoint-uri mw-api-tokens)
		    (puthash endpoint-uri (make-hash-table
					   :test 'equal)
			     mw-api-tokens))))
      (dolist (tok tokens)
	(when (funcall predicate tok)
	  (puthash (car tok) (cdr tok) hash))))))

(defun mw-api-response-check-http nil
  (cond ((not (= 200 url-http-response-status))
	 ;; FIXME: report also the reason phrase (per RFC 2068 6.1.1)?
	 (error "Unexpected HTTP response: %d"
		url-http-response-status))
	((not (string= "application/json"
		       (car (rfc2231-parse-string
			     url-http-content-type t))))
	 (error "Unexpected Content-Type: %s"
		url-http-content-type))))

(defun mw-api-response-check-json (json endpoint-uri)
  (dolist (elt json)
    (cond ((eq 'error (car elt))
	   ;; FIXME: we ought to be able to discern errors under
	   ;;        condition-case
	   (error "MediaWiki API error: %S" (cdr elt)))
	  ((eq 'warnings (car elt))
	   (dolist (w (cdr elt))
	     (display-warning (list 'mw-api (car w))
			      (format "MediaWiki: %S" (cdr w)))))
	  ((eq 'tokens (car elt))
	   (mw-api-store-tokens endpoint-uri (cdr elt)))
	  ((eq 'query  (car elt))
	   (let ((tokens (mw-api-get '(tokens) (cdr elt))))
	     (mw-api-store-tokens endpoint-uri
				  (or tokens
				      (mw-api-get '(pages nil) (cdr elt)))
				  (and (not tokens)
				       (lambda (pair)
					 (string-match-p
					  "token$"
					  (format "%s" (car pair)))))))))))

(defun mw-api-response-parse (endpoint-uri)
  (mw-api-response-check-http)
  (goto-char url-http-end-of-headers)
  (let ((json (json-read)))
    (mw-api-response-check-json json endpoint-uri)
    ;; .
    json))

(defun mw-api-call (endpoint-uri action parameters)
  "Perform a MediaWiki API call.
Issue a synchronous MediaWiki API request to the ENDPOINT-URI specified
and return the MediaWiki response, as an associative list, or signal an
error.

The request will use ACTION and \"json\" for the \"action\" and
\"format\" API parameters, respectively, and also any other parameters
passed as the PARAMETERS associative list.  If successful, the body of
the result will then be passed through `json-read'."
  (let ((url-request-method "POST")	; FIXME: allow for GET?
	(url-request-extra-headers
	 '(("Content-Type" . "application/x-www-form-urlencoded")))
	(url-request-data
	 (mm-url-encode-www-form-urlencoded
	  (nconc (list (cons "action" (mw-api-stringify action))
		       '("format" . "json"))
		 parameters)))
	buffer)
    (setq buffer (url-retrieve-synchronously endpoint-uri))
    ;; .
    (when buffer
      (unwind-protect
	  (with-current-buffer buffer
	    (mw-api-response-parse endpoint-uri))
	(kill-buffer buffer)))))

;;; Accessing saved tokens

(defun mw-api-token (endpoint-uri token)
  (let ((hash (gethash endpoint-uri mw-api-tokens)))
    ;; .
    (and hash
	 (gethash token hash))))

;;; Convenience

(defun mw-api-concat (a b)
  (concat a "|" b))

(defun mw-api-stringify (object)
  "Return a string representation of OBJECT.
If OBJECT is a string, return OBJECT.
If OBJECT is neither a string nor a list, return the printed
representation of OBJECT without quoting, as per `princ'.
If OBJECT is a list, return the concatenation of the printed
representations of its individual elements, delimited by \"|\"."
  (cond ((stringp object)   object)
	((listp   object)
	 (mapconcat (lambda (elt)
		      (if (stringp elt) elt (format "%s" elt)))
		    object "|"))
	(t
	 (format "%s" object))))

(defun mw-api-replace-once (key merge new alist)
  (let ((tail alist)
	r)
    (while tail
      (if (equal key (caar tail))
	  (setq r (nconc (cl-ldiff alist tail)
			 (list (cons key (funcall merge new (cdar tail))))
			 (cdr tail))
		tail nil)
	(setq   tail (cdr tail))))
    ;; .
    (or r
	(cons (cons key new) alist))))

(defun mw-api-adjoin-unsafe (key value alist)
  (mw-api-replace-once key
		       'mw-api-concat
		       (mw-api-stringify value)
		       alist))

(defun mw-api-adjoin (key value alist)
  "Return ALIST with an association of KEY to VALUE adjoined.
If VALUE is an empty string or nil, return ALIST.
If ALIST has no pairs with a key equal to KEY, return ALIST with a
(KEY . VALUE) pair prepended to it.
Otherwise, non-destructively replace the existing pair for the given KEY
with one holding VALUE prepended to the value of that pair, using \"|\"
as the delimiter.

VALUE will be passed through `mw-api-stringify'.

This function does not guarantee any particular order of pairs in the
resulting associative list."
  ;; .
  (if (and value (not (equal "" value)))
      (mw-api-adjoin-unsafe key value alist)
    alist))

(provide 'mw-api)
;;; mw-api.el ends here
