;;; mw-buffer.el --- Tying buffers to MediaWiki pages  -*- lexical-binding: t; -*-

;; Copyright (C) 2014 Free Software Foundation, Inc.

;; Author: Ivan Shmakov <ivan@siamics.net>
;; Version: 0.1
;; Keywords: web-services, comm

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;; This code provides support for several buffer-local variables which
;; associate Emacs buffers with MediaWiki instances and pages.  It is
;; used by mw-auth, mw-info, and vc-mw.

;;; Code:
(require 'mw-api)

(defvar mw-buffer-associate-page-hook nil
  "List of functions to call after a page has been associated with a buffer.
See `mw-buffer-associate-page'.")

(defvar mw-buffer-endpoint-uri nil)
(defvar mw-buffer-page-id nil)
(defvar mw-buffer-page-revision nil)
(defvar mw-buffer-page-title nil)

(make-variable-buffer-local 'mw-buffer-endpoint-uri)
(make-variable-buffer-local 'mw-buffer-page-id)
(make-variable-buffer-local 'mw-buffer-page-revision)
(make-variable-buffer-local 'mw-buffer-page-title)

;; FIXME: consider mw-buffer-endpoint-uri safe if a member of a trusted list?
; (put 'mw-buffer-endpoint-uri  'safe-local-variable 'stringp)
;; actually, we also allow for string -id and -revision
;;;###autoload
(put 'mw-buffer-page-id       'safe-local-variable 'integerp)
;;;###autoload
(put 'mw-buffer-page-revision 'safe-local-variable 'integerp)
;;;###autoload
(put 'mw-buffer-page-title    'safe-local-variable 'stringp)

(defun mw-buffer-page-query (&optional alist ignore-rev-id-p)
  "Return a MediaWiki query for the page associated with the current buffer.
This function will return an associative list of a single pair having
\"revids\", \"pageids\", or \"titles\" (in the order of preference) as
the key and holding the value of the respective buffer-local variable.

Additional API parameters are specified with the ALIST argument.  If
ALIST already contains one of the parameters above, its value is
merged, non-destructively, with the one found for the current buffer.

If IGNORE-REV-ID-P is non-nil, the value of `mw-buffer-page-revision'
for the current buffer is not considered."
  ;; FIXME: any way to check for consistency of the values?
  ;; .
  (cond ((and (not ignore-rev-id-p) mw-buffer-page-revision)
	 (mw-api-adjoin "revids"  mw-buffer-page-revision alist))
	(mw-buffer-page-id
	 (mw-api-adjoin "pageids" mw-buffer-page-id alist))
	(mw-buffer-page-title
	 (mw-api-adjoin "titles"  mw-buffer-page-title alist))))

;;;###autoload
(defun mw-buffer-associate-endpoint (endpoint-uri)
  "Associate current buffer with the MediaWiki endpoint URI given.
Return ENDPOINT-URI."
  (interactive "sURI: ")
  ;; .
  (setq mw-buffer-endpoint-uri endpoint-uri))

(defun mw-buffer-associate-page-1 (alist &optional endpoint-uri)
  (let* ((res   (mw-api-call (or endpoint-uri mw-buffer-endpoint-uri)
			     "query" alist))
	 (page  (mw-api-get '(query pages t) res))
	 (rev   (cdr (assoc 'lastrevid page)))
	 (id    (cdr (assoc 'pageid page)))
	 (title (cdr (assoc 'title  page))))
    (if (or (not page)
	    (and (not (assoc 'missing page))
		 (or (not rev) (not id) (not title))))
	(error "Unexpected MediaWiki response: %S" res)
      (if (assoc 'missing page)
	  nil				; no such page
	(setq mw-buffer-page-id id
	      mw-buffer-page-revision rev
	      mw-buffer-page-title  title)
	(run-hooks 'mw-buffer-associate-page-hook)
	;; .
	page))))

;;;###autoload
(defun mw-buffer-associate-page (page &optional revision alist)
  "Associate current buffer with a specific MediaWiki page.
Return page metadata as an associative list, or nil if no page is
found.  If non-nil, the list is guaranteed to contain at least the
'lastrevid, 'pageid and 'title associations.

This function runs `mw-buffer-associate-page-hook' if any
association is actually performed.

The PAGE argument may be either a page identifier (itself either an
integer or a string representing one) or a page title.

If PAGE is a string, try to locate a page with title matching the
string.  If the string matches ^[0-9]+$, also try to locate a
page with a matching page identifier.  If no page is found,
associate the buffer with PAGE as a page title and return nil.

Alternatively, a specific REVISION (an integer or a string
representation of one) to associate the buffer with may be given.
The PAGE argument is then ignored.

When used interactively, a prefix argument, if numeric, gives the
revision.  With a non-numeric prefix argument, prompt user for
the revision instead.

This function requires the `mw-buffer-endpoint-uri' variable being
already set up.  See `mw-buffer-associate-endpoint'."
  (interactive
   (cond ((not current-prefix-arg)
	  (list (read-string (concat "MediaWiki page title or id"
				     (when mw-buffer-page-title
				       (format " (default %s)"
					       mw-buffer-page-title))
				     ": ")
			     mw-buffer-page-title)))
	 ((numberp  current-prefix-arg)
	  (list nil current-prefix-arg))
	 (t
	  (list nil
		(read-number "MediaWiki revision: "
			     mw-buffer-page-revision)))))
  (let ((tail (mw-api-adjoin "prop" "info" alist)))
    (if (or revision (integerp page))
	(mw-buffer-associate-page-1
	 (if revision
	     (mw-api-adjoin "revids" revision tail)
	   (mw-api-adjoin  "pageids" page     tail)))
      (or (mw-buffer-associate-page-1
	   (mw-api-adjoin "titles" page tail))
	  ;; check if it is a string form of a revision number
	  (if (string-match-p "^[0-9]+$" page)
	      (mw-buffer-associate-page-1
	       (mw-api-adjoin "pageids"  page tail))
	    ;; assume it is a title for a new page
	    (setq mw-buffer-page-title page)
	    (run-hooks 'mw-buffer-associate-page-hook)
	    ;; .
	    nil)))))

;;; Desktop support
(add-to-list 'desktop-locals-to-save 'mw-buffer-endpoint-uri)
(add-to-list 'desktop-locals-to-save 'mw-buffer-page-id)
(add-to-list 'desktop-locals-to-save 'mw-buffer-page-revision)
(add-to-list 'desktop-locals-to-save 'mw-buffer-page-title)

(provide 'mw-buffer)
;;; mw-buffer.el ends here
