;;; mw-eww.el --- Previewing MediaWiki markup with EWW  -*- lexical-binding: t; -*-

;; Copyright (C) 2014 Free Software Foundation, Inc.

;; Author: Ivan Shmakov <ivan@siamics.net>
;; Version: 0.1
;; Keywords: web-services, html, comm

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;; The mw-eww-preview command (the only one provided by this module)
;; passes the contents of the current buffer through the associated
;; MediaWiki instance (per mw-buffer-endpoint-uri) and renders the
;; resulting HTML in the EWW buffer referenced by the
;; mw-eww-preview-buffer variable (creating one if nil).

;; Please note that this code relies on EWW multi-buffer support, as
;; suggested in http://debbugs.gnu.org/16211.

;;; Code:
(require 'eww)
(require 'mw-api)
(require 'mw-buffer)

(defvar mw-eww-preview-buffer nil
  "EWW buffer associated with the current buffer.")

(make-variable-buffer-local 'mw-eww-preview-buffer)

(defun mw-eww-preview (&optional markup endpoint-uri alist)
  "Preview MARKUP as a MediaWiki page via ENDPOINT-URI.
If MARKUP is nil, the contents of the current buffer is used instead."
  (interactive
   (list (buffer-substring-no-properties
	  (point-min) (point-max))
	 ;; FIXME: refactor this part?
         (or (and (not current-prefix-arg)
                  mw-buffer-endpoint-uri)
             (call-interactively
              'mw-buffer-associate-endpoint))))
  (unless endpoint-uri
    (setq endpoint-uri mw-buffer-endpoint-uri))
  (let* ((result
	  (mw-api-call
	   endpoint-uri "parse"
	   `(("text" . ,markup)
             ("disablepp" . "1")
             ("pst" . "1")
             ,@(mw-api-adjoin
                "prop" '(text categorieshtml languageshtml)
                (if mw-buffer-page-title
                    (mw-api-adjoin "title" mw-buffer-page-title
                                   alist)
                  alist)))))
	 (html  (mw-api-get '(parse text *) result))
         (title mw-buffer-page-title)
         (endpt endpoint-uri)
	 buffer)
    (unless html
      (error "Cannot render wikitext given; details: %S" result))
    (unless mw-eww-preview-buffer
      (setq mw-eww-preview-buffer
            (generate-new-buffer
             (concat "*MediaWiki preview of "
                     (buffer-name) "*"))))
    (with-current-buffer mw-eww-preview-buffer
      (if (or eww-current-dom
	      eww-current-source)
	  (eww-save-history)
	(eww-mode)))
    (setq buffer mw-eww-preview-buffer)
    (with-temp-buffer
      (insert "<title >"
	      (xml-escape-string title)
	      " "
	      (xml-escape-string endpt)
	      "</title>")
      (insert html)
      (let ((cats  (mw-api-get '(parse categorieshtml *) result))
	    (langs (mw-api-get '(parse languageshtml  *) result)))
	(when cats   (insert cats))
	(when langs  (insert langs)))
      (goto-char (point-min))
      (setq-local eww-current-buffer  buffer)
      ;; FIXME: utf8 inhibits conversion for eww-display-html
      (eww-display-html 'utf8 nil))
    (pop-to-buffer mw-eww-preview-buffer)))

(provide 'mw-eww)
;;; mw-eww.el ends here
