;;; vc-mw.el --- Support for MediaWiki as a vc.el backend  -*- lexical-binding: t; -*-

;; Copyright (C) 2014 Free Software Foundation, Inc.

;; Author: Ivan Shmakov <ivan@siamics.net>
;; Version: 0.1
;; Keywords: vc, web-services, comm

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;; This backend is unusual in that it uses buffer-local variables to
;; hold the relevant state.  These may either be set via
;; .dir-locals.el (mw-buffer-endpoint-uri), or preserved via
;; desktop.el.

;; Unfortunately, as vc.el facilities generally operate on files (and
;; not buffers), it becomes necessary to rely on files (and
;; buffer-file-name), even though this backend could have been made to
;; actually work without ever accessing the filesystem.

;; As opposed to the majority of other vc.el backends, this one does
;; not require any external software to operate.

;;; Code:

(eval-when-compile
  (require 'vc))

(require 'mw-api)
(require 'mw-buffer)

(defgroup vc-mw nil
  "MediaWiki vc.el backend."
  :group 'vc)

(defcustom vc-mw-file-string 'vc-mw-buffer-file-string
  "The function to use to get the contents of the associated file."
  :type '(choice (function-item :tag "File contents"
				vc-mw-buffer-file-string)
		 (function-item :tag "Buffer contents, respect narrowing"
				buffer-string)
		 (function-item :tag "Buffer contents, no narrowing"
				vc-mw-buffer-string-wide)
		 (function :tag "Custom"))
  :group 'vc-mw)

(defcustom vc-mw-revision-prefix "r"
  "Prefix to put before the revision number for the purposes of vc.el."
  :type 'string
  :group 'vc-mw)

(defvar vc-mw-buffer-time-stamp nil
  "Timestamp for the last interaction with the server.")
(defvar vc-mw-buffer-page-time-stamp nil
  "Timestamp for the latest revision associated with the buffer.")

(make-variable-buffer-local 'vc-mw-buffer-time-stamp)
(make-variable-buffer-local 'vc-mw-buffer-page-time-stamp)

(defun vc-mw-unprefix-revision (revision)
  (let ((len (length vc-mw-revision-prefix)))
    ;; .
    (if (and revision
	     (> (length revision) len)
	     (string= (substring revision 0 len)
		      vc-mw-revision-prefix))
	(substring revision len)
      revision)))

(defun vc-mw-log-queries (files &optional shortlog start-revision limit alist)
  ;; FIXME: implement shortlog handling?
  (setq alist (mw-api-adjoin "rvlimit" (or limit "max")
			     (mw-api-adjoin "prop" "revisions" alist)))
  (when start-revision
    (setq alist (mw-api-adjoin
		 "rvstartid" (vc-mw-unprefix-revision start-revision)
		 alist)))
  (let ((result nil))
    (dolist (file files)
      (let ((buffer (vc-mw-buffer-for-file file)))
	(when buffer
	  (with-current-buffer buffer
	    (let* ((uri mw-buffer-endpoint-uri)
		   (pair  (and (equal 1 limit)  ; NB: API restriction
			       (assoc uri result)))
		   (new (mw-buffer-page-query (or (cdr pair) alist)
					      'ignore-rev-id)))
	      (cond ((not new)
		     (warn
		      "%s (%s): File is not under MediaWiki version control"
		      file buffer))
		    (pair (setcdr pair new))
		    (t    (setq result (cons (cons uri new) result)))))))))
    ;; .
    result))

(defun vc-mw-print-revisions (revisions &optional buffer)
  (save-current-buffer
    (if buffer
	(set-buffer buffer))
    (let ((i 0))
      (while (< i (length revisions))
	(insert "\n")
	(dolist (pair (aref revisions i))
	  (unless (eq '* (car pair))
	    ;; FIXME: improve formatting?
	    (insert (format "%s: %s\n" (car pair) (cdr pair)))))
	(setq i (+ 1 i))))))

;;;###autoload
(defun vc-mw-buffer-for-file (file)
  ;; .
  (find-buffer-visiting
   file
   (lambda (buffer)
     (with-current-buffer buffer
       (and (boundp 'mw-buffer-endpoint-uri)	; may get here early
	    mw-buffer-endpoint-uri
	    (or mw-buffer-page-id
		mw-buffer-page-revision
		mw-buffer-page-title))))))

(defun vc-mw-buffer-file-string nil
  (let ((file (buffer-file-name)))
    (with-temp-buffer
      (insert-file-contents file)
      (buffer-string))))

(defun vc-mw-buffer-string-wide nil
  (save-restriction
    (widen)
    (buffer-string)))

;;; MediaWiki API calls

(defun vc-mw-api-call (endpoint-uri action &optional parameters)
  (with-demoted-errors "MediaWiki API: %S"
    (mw-api-call endpoint-uri action parameters)))

(defun vc-mw-buffer-page-info (&optional buffer alist)
  (save-current-buffer
    (if buffer
	(set-buffer buffer))
    (when mw-buffer-endpoint-uri
      (let* ((par-1 (mw-buffer-page-query alist))
	     (params
	      (and par-1
		   (mw-api-adjoin
		    "prop" '(info revisions)
		    (mw-api-adjoin "rvprop" "timestamp" par-1))))
	     (res
	      (and params
		   (vc-mw-api-call mw-buffer-endpoint-uri
				   "query"
				   (if (mw-api-token mw-buffer-endpoint-uri
						     'csrftoken)
				       params
				     ;; request an edit token, just in case
				     (mw-api-adjoin "intoken" "edit"
						    params)))))
	     (page  (mw-api-get '(query pages t) res))
	     (revs  (cdr (assoc 'revisions page)))
	     (rtop  (and revs (< 0 (length revs))
			 (aref revs 0)))
	     (stamp (cdr (assoc 'timestamp rtop))))
	(when rtop
	  (setq vc-mw-buffer-time-stamp
		(or (cdr (assoc 'starttimestamp rtop))
		    ;; FIXME: huh?
		    (format-time-string "%Y-%m-%dT%H:%M:%SZ" nil t)))
	  (if (or (not     vc-mw-buffer-page-time-stamp)
		  (string< vc-mw-buffer-page-time-stamp stamp))
	      (setq vc-mw-buffer-page-time-stamp stamp)))
	;; .
	page))))

(defun vc-mw-buffer-state (&optional buffer)
  (let* ((params '(("prop" . "revisions")))
	 (page  (cdr
		 (vc-mw-buffer-page-info
		  buffer
		  ;; also get the content if the revision is known
		  (if mw-buffer-page-revision
		      (mw-api-adjoin "rvprop" "content"
				     params)
		    params))))
	 (head  (cdr (assoc 'lastrevid page)))
	 (revs  (cdr (assoc 'revisions page)))
	 (data  (and revs (cdr (assoc '* (aref revs 0)))))
	 (samep (and data
		     (string= (funcall vc-mw-file-string)
			      (concat data "\n")))))
    ;; .
    (cond ((not page)
	   'unregistered)
	  ((assoc 'missing page)
	   'added)
	  ((= head mw-buffer-page-revision)
	   (if samep 'up-to-date 'edited))
	  (t
	   (if samep 'needs-update 'needs-merge)))))

(defun vc-mw-buffer-associated (&optional buffer)
  ;; FIXME: looking for a better name; and a documentation
  (interactive)
  (vc-file-clearprops (buffer-file-name buffer))
  (vc-mw-buffer-state buffer))

;; update state upon associating a buffer to a page
(add-hook 'mw-buffer-associate-page-hook
	  'vc-mw-buffer-associated)

(defun vc-mw-buffer-checkin (comment &optional buffer alist)
  (let* ((e-uri mw-buffer-endpoint-uri)
	 (text  (let ((c (funcall vc-mw-file-string)))
		  ;; suppress a trailing newline (see vc-mw-find-revision)
		  (if (= ?\n (aref c (+ -1 (length c))))
		      (substring c 0 -1)
		    c)))
	 (params
	  (nconc (list (cons "title"    mw-buffer-page-title)
		       (cons "text"     text)
		       (cons "summary"  comment)
		       (cons "basetimestamp"
			     (or vc-mw-buffer-page-time-stamp ""))
		       (cons "starttimestamp"
			     (or vc-mw-buffer-time-stamp ""))
		       (cons "token"
			     (or (mw-api-token e-uri 'csrftoken)
				 (mw-api-token e-uri 'edittoken)
				 "")))
		 alist))
	 (response  (mw-api-call e-uri "edit" params))
	 (edit  (cdr (assoc 'edit   response)))
	 (outc  (cdr (assoc 'result edit))))
    (unless (equal "Success" outc)
      (error "Unexpected MediaWiki response: %S" response))
    ;; .
    response))

;;; vc.el interface

(defun vc-mw-revision-granularity nil
  'file)

;; This function will be copied directly into loaddefs.el, so that
;; vc-mw.el loading can be deferred until actually needed.  Please
;; note that it also requires vc-mw-buffer-for-file above.
;;;###autoload
(progn
(defun vc-mw-registered (file)
  (let ((buffer (vc-mw-buffer-for-file file)))
    (when buffer
      ;; .
      (vc-mw-buffer-page-info buffer)))))

(defun vc-mw-state (file)
  (let ((buffer (vc-mw-buffer-for-file file)))
    (if buffer
	(with-current-buffer buffer
	  (vc-mw-buffer-state))
      'unregistered)))

(defun vc-mw-working-revision (file)
  (let ((buffer (vc-mw-buffer-for-file file)))
    (when buffer
      (with-current-buffer buffer
	;; vc.el requires a string here, ensure it gets one
	(format "%s%s" (or vc-mw-revision-prefix "")
		(or mw-buffer-page-revision 0))))))

(defun vc-mw-checkout-model (files)
  ;; FIXME: huh?
  'implicit)

(defun vc-mw-find-revision (file rev buffer)
  (setq rev (vc-mw-unprefix-revision rev))
  (save-current-buffer
    (set-buffer (vc-mw-buffer-for-file file))
    (let* ((par-1 '(("prop"   . "revisions")
		    ("rvprop" . "content")))
	   (params
	    (cond ((not rev)
		   (mw-buffer-page-query par-1))
		  ((string= "" rev)
		   (mw-buffer-page-query par-1 'ignore-rev-id))
		  (t
		   (mw-api-adjoin "revids" rev par-1))))
	   (result
	    ;; NB: allowing for exceptions here
	    (mw-api-call mw-buffer-endpoint-uri "query" params))
	   (page    (mw-api-get '(query pages t) result))
	   (revs    (cdr (assoc 'revisions page)))
	   (content (and revs (cdr (assoc '* (aref revs 0))))))
      (set-buffer buffer)
      (when content
	(insert content)
	;; ensure terminating newline (for diff purposes)
	(unless (= ?\n (char-before (point)))
	  (insert "\n"))))))

(defun vc-mw-checkin (files rev comment)
  (save-current-buffer
    (dolist (file files)
      (set-buffer (vc-mw-buffer-for-file file))
      (vc-mw-buffer-checkin comment))))

(defun vc-mw-print-log (files buffer &optional shortlog start-revision limit)
  (let ((queries (vc-mw-log-queries files shortlog start-revision limit)))
    (save-current-buffer
      (vc-setup-buffer buffer)		; does set-buffer for us
      (dolist (pair queries)
	(let* ((inhibit-read-only t)
	       (uri (car pair))
	       (r   (mw-api-call uri "query" (cdr pair)))
	       (q   (cdr (assoc 'query r))))
	  (when (cdr queries)
	    (insert "\n*** " uri "\n"))
	  (dolist (page (cdr (assoc 'pages q)))
	    (vc-mw-print-revisions
	     (cdr (assoc 'revisions (cdr page))))))))))

;;; unload-feature support

(defun vc-mw-unload-function nil
  (remove-hook 'mw-buffer-associate-page-hook
	       'vc-mw-buffer-associated)
  ;; .
  nil)

(provide 'vc-mw)
;;; vc-mw.el ends here
